(ns rlm.qotd)

;;; There is a pair of integers, 1 < m < n, whose sum is
;;; less than 100.

;; These constraints makes a triangular sort of pattern.
(defn generate-valid-numbers
  [n]
  (filter (fn [[a b]]  (< a b))
          (map #(vector % n) (range 2 (- 150 n)))))

(def squish (partial reduce concat))

(def p0 (squish (map generate-valid-numbers (range 2 149))))

;;; Person S knows their sum, but nothing else about them.
;;; Person P knows their product, but nothing else about
;;; them.

;;; Now, Persons S and P know the above information, and
;;; each one knows that the other one knows it. They have
;;; the following conversation:

;;; P: I can't figure out what the numbers are.

;; Eliminate pairs with a unique product.

(defn group-by
  "Split coll into groups where the f maps each element in a
  group to the same value in O(n*log(n)) time."
  [f coll]
  (partition-by f (sort-by f coll)))

(defn unique-by
  "Remove all elements a,b of coll that for which
   (= (f a) (f b)) in O(n*log(n)) time."
  [f coll]
  (squish (filter #(= (count %) 1) (group-by f coll))))
  
(defn multiple-by
  "Keep all elements a,b of coll for which
   (= (f a) (f b)) in O(n*log(n)) time."
  [f coll]
  (squish (filter #(> (count %) 1) (group-by f coll))))

(defn prod [[a b]] (* a b))

(def p1 (multiple-by prod p0))

;;; S: I was sure you couldn't.

;; Each possible sum s has a set of possible pairs [a b]
;; where (= s (+ a b)). Partition p0 (since S *was* sure) by
;; sum, and keep those pairs that belong in partitions where
;; each pair in that partition is in p1. (since the only way
;; he could be *sure*, is if every possibility for a given
;; sum had an ambiguous product.

(defn sum [[a b]] (+ a b))

(def p2
  (squish
   (filter
    (partial every? (set p1))
    (group-by sum p0))))


;;; P: Then I have.

;; Keep those pairs that have a unique product out of the
;; ones that are left.

(def p3
  (unique-by prod p2))


;;; S: Then I have, too.

;; Keep those pairs that have a unique sum out of the
;; ones that are left.

(def p4 (unique-by sum p3))


(defn solve [limit]
  (let [generate-valid-numbers
        (fn 
          [n]
          (filter (fn [[a b]]  (< a b))
                  (map #(vector % n)
                       (range 2 (- limit n)))))
        p0
        (squish (map generate-valid-numbers
                     (range 2 (dec limit))))
        p2
        (squish
         (filter
          (partial every? (set p1))
            (group-by sum p0)))

        p3 (unique-by prod p2)]
    (unique-by sum p3)))

;;; Dylan START!
(defn results [beez]
  (let
      [
       pairs
       (for [m (range 2 beez)
             n (range (inc m) beez)]
         [m n])
       
       singleton? (comp zero? dec count)

       sum (fn [[x y]] (+ x y))
       product (fn [[x y]] (* x y))

       inverse-sum
       (fn [s]
         (filter #(= s (sum %)) pairs))
       
       inverse-product
       (fn [p]
         (filter #(= p (product %)) pairs))
       ] 

    
    (->>
     pairs
     
     (filter #(< (sum %) beez))

     ;; P: I cannot find the numbers.
     (remove (comp singleton?
                   inverse-product
                   product))
     set
     
     ;; S: I was sure you couldn't.
     ((fn [coll]
        (filter
         (comp (partial every? coll)
               inverse-sum
               sum)
         coll)))
     set

     ;;P: Then I have.
     ((fn [coll]
        (filter
         (comp singleton?
               (partial filter coll)
               inverse-product
               product)
         coll)))
     set
     
     ;;S: Then I have, too.
     ((fn [coll]
        (filter
         (comp singleton?
               (partial filter coll)
               inverse-sum
               sum)
         coll)))

     )))